"""The applicaton."""
from abc import ABC, abstractmethod
import uuid


class Optimizer(ABC):
    """Base class for optimizers."""

    @abstractmethod
    def get_optimized_power(self, deviceId: uuid.UUID) -> float:
        """Get the optimized power budget for a give device."""
        pass



"""Main module for the energy assistant application."""
import asyncio
from contextlib import asynccontextmanager, suppress
from datetime import date
import json
import logging
from logging.handlers import RotatingFileHandler
import os
import sys
import threading
from typing import AsyncIterator, Final

from apscheduler.schedulers.asyncio import AsyncIOScheduler  # type: ignore
from colorlog import ColoredFormatter
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from fastapi_socketio import SocketManager  # type: ignore
import requests  # type: ignore
from sqlalchemy.ext.asyncio import AsyncSession
import yaml

from app.EmhassOptimizer import EmhassOptimizer
from app.api.device import OTHER_DEVICE
from app.api.main import router as api_router
from app.devices import StatesMultipleRepositories, StatesRepository
from app.devices.device import Device
from app.devices.evcc import EvccDevice
from app.devices.home import Home
from app.devices.homeassistant import Homeassistant
from app.devices.registry import DeviceTypeRegistry
from app.mqtt import MqttConnection
from app.settings import settings
from app.storage import Database, get_async_session, session_storage

from .constants import ROOT_LOGGER_NAME

FORMAT_DATE: Final = "%Y-%m-%d"
FORMAT_TIME: Final = "%H:%M:%S"
FORMAT_DATETIME: Final = f"{FORMAT_DATE} {FORMAT_TIME}"
MAX_LOG_FILESIZE = 1000000 * 10  # 10 MB


class EnergyAssistant:
    """Energy Assistant Application."""

    home: Home
    hass: Homeassistant | None
    optimizer: EmhassOptimizer
    mqtt: MqttConnection | None
    db: Database


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncIterator:
    """Manage the startup and showdown."""
    ea = await init_app()
    sio.start_background_task(background_task, ea)
    sio.start_background_task(optimize, ea.optimizer)

    scheduler = AsyncIOScheduler()
    scheduler.add_job(
        daily_optimize, trigger="cron", args=[ea], hour="3", minute="0"
    )  # time is UTC
    scheduler.start()
    yield
    scheduler.shutdown()
    print("Shutdown app")


app = FastAPI(title="energy-assistant", lifespan=lifespan)
sio = SocketManager(app=app, cors_allowed_origins="*")

origins = [
    "http://localhost",
    "http://localhost:5000",
    "http://localhost:5173",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(api_router, prefix="/api")


async def async_handle_state_update(
    ea: EnergyAssistant,
    state_repository: StatesRepository,
    session: AsyncSession,
) -> None:
    """Read the values from home assistant and process the update."""
    try:
        state_repository.read_states()

        await ea.home.update_state(state_repository)
        if ea.optimizer is not None:
            await ea.home.update_power_consumption(state_repository, ea.optimizer)
            ea.optimizer.update_repository_states(ea.home, state_repository)
        else:
            logging.error("The variable optimizer is None in async_handle_state_update")
        state_repository.write_states()
        # print("Send refresh: " + get_home_message(home))
        if ea.db:
            if ea.home:
                await asyncio.gather(
                    sio.emit("refresh", {"data": get_home_message(ea.home)}),
                    ea.db.store_home_state(ea.home, session),
                )
            else:
                logging.error("The variable home is None in async_handle_state_update")
        else:
            logging.error("The variable db is None in async_handle_state_update")
        if ea.optimizer is not None and ea.home is not None:
            ea.optimizer.update_devices(ea.home)
    except Exception:
        logging.exception("error during sending refresh")


async def background_task(ea: EnergyAssistant) -> None:
    """Periodically read the values from home assistant and process the update."""
    last_update = date.today()
    async_session = await get_async_session()
    repositories: list[StatesRepository] = []
    if ea.hass is not None:
        repositories.append(ea.hass)
    if ea.mqtt is not None:
        repositories.append(ea.mqtt)
    state_repository = StatesMultipleRepositories(repositories)
    while True:
        await sio.sleep(30)
        # delta_t = datetime.now().timestamp()
        # print("Start refresh from home assistant")
        today = date.today()
        try:
            if today != last_update:
                ea.home.store_energy_snapshot()
            last_update = today
            async with async_session() as session:
                await async_handle_state_update(ea, state_repository, session)
        except Exception:
            logging.exception("error in the background task")
        # print(f"refresh from home assistant completed in {datetime.now().timestamp() - delta_t} s")


def get_self_sufficiency(consumed_solar_energy: float, consumed_energy: float) -> float:
    """Calulate the self sufficiency value."""
    return min(
        round(consumed_solar_energy / consumed_energy * 100) if consumed_energy > 0 else 0.0, 100
    )


def get_device_message(device: Device) -> dict:
    """Generate the update data message for a device."""
    if device.energy_snapshot is not None:
        consumed_energy_today = device.consumed_energy - device.energy_snapshot.consumed_energy
        consumed_solar_energy_today = (
            device.consumed_solar_energy - device.energy_snapshot.consumed_solar_energy
        )
    else:
        consumed_energy_today = 0
        consumed_solar_energy_today = 0
    result = {
        "device_id": str(device.id),
        "type": device.__class__.__name__,
        "power": device.power,
        "available": device.available,
        "today": {
            "consumed_solar_energy": consumed_solar_energy_today,
            "consumed_energy": consumed_energy_today,
            "self_sufficiency": get_self_sufficiency(
                consumed_solar_energy_today, consumed_energy_today
            ),
        },
    }
    result["attributes"] = device.attributes
    return result


def get_home_message(home: Home) -> str:
    """Generate the update data message for a home."""
    devices_messages = []
    if home.energy_snapshop is not None:
        consumed_energy_today = home.consumed_energy - home.energy_snapshop.consumed_energy
        consumed_solar_energy_today = (
            home.consumed_solar_energy - home.energy_snapshop.consumed_solar_energy
        )
    else:
        consumed_energy_today = 0
        consumed_solar_energy_today = 0

    other_power = home.home_consumption_power
    other_consumed_energy = consumed_energy_today
    other_consumed_solar_energy = consumed_solar_energy_today
    for device in home.devices:
        device_message = get_device_message(device)
        devices_messages.append(device_message)
        other_power = other_power - device.power
        other_consumed_energy = other_consumed_energy - device_message["today"]["consumed_energy"]
        other_consumed_solar_energy = (
            other_consumed_solar_energy - device_message["today"]["consumed_solar_energy"]
        )

    other_device = {
        "device_id": str(OTHER_DEVICE),
        "type": "other_device",
        "power": other_power,
        "available": True,
        "today": {
            "consumed_solar_energy": other_consumed_solar_energy,
            "consumed_energy": other_consumed_energy,
            "self_sufficiency": get_self_sufficiency(
                other_consumed_solar_energy, other_consumed_energy
            ),
        },
    }
    devices_messages.append(other_device)

    home_message = {
        "name": home.name,
        "power": {
            "solar_production": home.solar_production_power,
            "grid_supply": home.grid_imported_power,
            "solar_self_consumption": home.solar_self_consumption_power,
            "home_consumption": home.home_consumption_power,
            "self_sufficiency": round(home.self_sufficiency * 100),
        },
        "today": {
            "consumed_solar_energy": consumed_solar_energy_today,
            "consumed_energy": consumed_energy_today,
            "self_sufficiency": get_self_sufficiency(
                consumed_solar_energy_today, consumed_energy_today
            ),
        },
        "devices": devices_messages,
    }
    return json.dumps(home_message)


@app.sio.event  # type: ignore
async def connect(sid, environ):
    """Handle the connect of a client via socket.io to the server."""
    logging.info(f"connect {sid}")
    await sio.emit("refresh", {"data": get_home_message(app.home)}, room=sid)


@app.sio.event  # type: ignore
def disconnect(sid):
    """Handle the disconnect of a client via socket.io to the server."""
    logging.info(f"Client disconnected {sid}")


def create_mqtt_connection(config: dict) -> MqttConnection | None:
    """Create an mqtt connection based on the config."""

    mqtt_config = config.get("mqtt")
    if mqtt_config is not None:
        mqtt_host = mqtt_config.get("host")
        mqtt_username = mqtt_config.get("username")
        mqtt_password = mqtt_config.get("password")
        mqtt_topic = mqtt_config.get("topic")
        mqtt_connection = MqttConnection(mqtt_host, mqtt_username, mqtt_password, mqtt_topic)
        mqtt_connection.connect()
        return mqtt_connection
    return None


def subscribe_mqtt_topics(mqtt_connection: MqttConnection, home: Home) -> None:
    """Subscribe the mqtt based devices on the mqtt connection."""

    for device in home.devices:
        if isinstance(device, EvccDevice):
            mqtt_connection.add_subscription_topic(device.evcc_mqtt_subscription_topic)


def create_hass_connection(config: dict) -> Homeassistant | None:
    """Create a connection to home assistant."""
    token: str | None = None
    url: str | None = None
    try:
        token = os.getenv("SUPERVISOR_TOKEN")
        if token is not None:
            logging.info(f"suvervisor token detected. len={len(token)}")
            url = "http://supervisor/core"

            headers = {
                "Authorization": f"Bearer {token}",
                "content-type": "application/json",
            }
            response = requests.get(f"{url}/api/states", headers=headers)
            logging.info(
                f"pinging homeassistant api succeeeded. Status code = {response.status_code}"
            )
            if response.ok:
                logging.info(f"Using {url} to connect")
                hass = Homeassistant(url, token, False)
                return hass
    except Exception:
        logging.exception("Error while trying to connect to the homeassistant supervisor api")
        url = None
        token = None

    logging.info("Try to connect to home assistant based on the config file entries...")
    hass_config = config.get("homeassistant")
    if hass_config is not None:
        demo_mode = hass_config.get("demo_mode")
        url = hass_config.get("url")
        token = hass_config.get("token")
        if url is not None and token is not None:
            return Homeassistant(url, token, demo_mode)
    return None


def setup_logger(log_filename: str, level: str = "DEBUG") -> logging.Logger:
    """Initialize logger."""
    # define log formatter
    log_fmt = "%(asctime)s.%(msecs)03d %(levelname)s (%(threadName)s) [%(name)s] %(message)s"

    # base logging config for the root logger
    logging.basicConfig(level=logging.INFO)

    colorfmt = f"%(log_color)s{log_fmt}%(reset)s"
    logging.getLogger().handlers[0].setFormatter(
        ColoredFormatter(
            colorfmt,
            datefmt=FORMAT_DATETIME,
            reset=True,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red",
            },
        )
    )

    # Capture warnings.warn(...) and friends messages in logs.
    # The standard destination for them is stderr, which may end up unnoticed.
    # This way they're where other messages are, and can be filtered as usual.
    logging.captureWarnings(True)

    # setup file handler
    # log_filename = os.path.join(data_path, "energy_assistant.log")
    file_handler = RotatingFileHandler(log_filename, maxBytes=MAX_LOG_FILESIZE, backupCount=1)
    # rotate log at each start
    with suppress(OSError):
        file_handler.doRollover()
    file_handler.setFormatter(logging.Formatter(log_fmt, datefmt=FORMAT_DATETIME))
    # file_handler.setLevel(logging.INFO)

    logger = logging.getLogger()
    logger.addHandler(file_handler)

    # apply the configured global log level to the (root) music assistant logger
    logging.getLogger(ROOT_LOGGER_NAME).setLevel(level)

    # silence some noisy loggers
    logging.getLogger("asyncio").setLevel(logging.WARNING)
    logging.getLogger("aiosqlite").setLevel(logging.WARNING)
    logging.getLogger("databases").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)

    sys.excepthook = lambda *args: logging.getLogger(None).exception(
        "Uncaught exception",
        exc_info=args,  # type: ignore[arg-type]
    )
    threading.excepthook = lambda args: logging.getLogger(None).exception(
        "Uncaught thread exception",
        exc_info=(  # type: ignore[arg-type]
            args.exc_type,
            args.exc_value,
            args.exc_traceback,
        ),
    )

    return logger


async def init_app() -> EnergyAssistant:
    """Initialize the application."""
    result = EnergyAssistant()

    hass_options_file = "/data/options.json"
    if os.path.isfile(hass_options_file):
        with open(hass_options_file, "rb") as _file:
            hass_options = json.loads(_file.read())
    else:
        hass_options = {}

    log_level = hass_options.get("log_level", "info").upper()

    logger = setup_logger(settings.LOG_FILE, log_level)

    config_file = hass_options.get("config_file", settings.CONFIG_FILE)

    logger.info("Hello from Energy Assistant")

    for option in hass_options:
        logging.info(f"{option}={hass_options[option]}")

    async_session = await get_async_session()
    db = Database()
    result.db = db

    device_type_registry = DeviceTypeRegistry()
    device_type_registry.load(settings.DEVICE_TYPE_REGISTRY)

    logger.info(f"Loading config file {config_file}")
    try:
        with open(config_file) as stream:
            logger.debug(f"Successfully opened config file {config_file}")
            try:
                config = yaml.safe_load(stream)
                logger.debug(f"config file {config_file} successfully loaded")
            except yaml.YAMLError:
                logger.exception("Failed to parse the config file")
            except Exception:
                logger.exception("Failed to parse the config file")
            else:
                hass = create_hass_connection(config)
                result.hass = hass
                if hass is not None:
                    hass.read_states()
                    optimizer = EmhassOptimizer(settings.DATA_FOLDER, config, hass)
                    result.optimizer = optimizer
                    app.optimizer = optimizer  # type: ignore

                mqtt_connection: MqttConnection | None = create_mqtt_connection(config)
                result.mqtt = mqtt_connection

                home_config = config.get("home")
                if home_config is not None and home_config.get("name") is not None:
                    home = Home(home_config, session_storage, device_type_registry)
                    result.home = home
                    app.home = home  # type: ignore
                    if mqtt_connection is not None:
                        subscribe_mqtt_topics(mqtt_connection, home)

                    async with async_session() as session:
                        await db.update_devices(
                            home, session, session_storage, device_type_registry
                        )

                        await db.restore_home_state(home, session)
                else:
                    logger.error(f"home not found in config file: {config}")
                logger.info("Initialization completed")
    except Exception:
        logger.exception("Initialization of the app failed")
    return result


async def optimize(optimizer: EmhassOptimizer) -> None:
    """Optimize the forcast."""
    try:
        optimizer.forecast_model_fit(True)
    except Exception:
        logging.exception(
            "Optimization of the power consumption forcast model failed, probably due to missing history data in Home Assistant."
        )


def daily_optimize(ea: EnergyAssistant) -> None:
    """Optimze once a day."""
    try:
        optimizer = ea.optimizer
        if optimizer is not None:
            logging.info("Start optimizer run")
            optimizer.dayahead_forecast_optim()
    except Exception:
        logging.exception("Daily optimization run failed")


@app.get("/check", include_in_schema=False)
async def health() -> JSONResponse:
    """Test the web server with a ping."""
    return JSONResponse({"message": "It worked!!"})


# This needs to be the last mount
app.mount("/", StaticFiles(directory="client", html=True), name="frontend")


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=5000)



"""Tests for the devices."""
from datetime import date
import uuid

import pytest
from sqlalchemy.ext.asyncio import AsyncSession

from app.devices import EnergyIntegrator, PowerModes
from app.devices.home import Home
from app.devices.homeassistant import HomeassistantDevice
from app.devices.registry import DeviceTypeRegistry
from app.models.home import HomeMeasurement
from app.storage import Database, session_storage


async def setup_data(session: AsyncSession) -> None:
    """Set up the data in the database."""
    from app.models.device import Device, DeviceMeasurement
    from app.models.home import HomeMeasurement

    device1 = Device(
        id=uuid.UUID("1a8ac2d6-5695-427a-a3c5-ef567b34e5ec"),
        name="Device 1",
        icon="mdi-home",
        power_mode=PowerModes.DEVICE_CONTROLLED,
    )
    session.add(device1)
    device2 = Device(
        id=uuid.UUID("2a8ac2d6-5695-427a-a3c5-ef567b34e5ec"),
        name="Device 2",
        icon="mdi-home",
        power_mode=PowerModes.DEVICE_CONTROLLED,
    )
    session.add(device2)
    await session.flush()
    devices = [device1, device2]

    dates = [date(2023, 1, 9), date(2023, 1, 10), date(2023, 1, 11)]
    for m, d in enumerate(dates):
        home_measurement = HomeMeasurement(
            name="my home",
            measurement_date=d,
            consumed_energy=123 + m,
            solar_consumed_energy=120 + 0.5 * m,
            solar_produced_energy=150 + m,
            grid_imported_energy=1123 + m,
            grid_exported_energy=1540 + m,
            device_measurements=[],
        )
        session.add(home_measurement)
        await session.flush()
        for i, device in enumerate(devices):
            device_measurement = DeviceMeasurement(
                home_measurement_id=home_measurement.id,
                consumed_energy=2 + 0.5 * i + m,
                solar_consumed_energy=1 + i + 0.5 * m,
                device_id=device.id,
            )
            session.add(device_measurement)
        await session.flush()

    await session.commit()


def test_integrator() -> None:
    """Test the energy integrator."""
    integrator = EnergyIntegrator()
    integrator.restore_state(15, 20)
    integrator.add_measurement(30, 0.1)
    assert integrator.consumed_solar_energy == 16


@pytest.mark.asyncio
async def test_load(session: AsyncSession, device_type_registry: DeviceTypeRegistry) -> None:
    """Test the loading of the devices."""
    await setup_data(session)

    home = Home(
        {
            "name": "my home",
            "solar_power": "solar_power_id",
            "grid_supply_power": "grid_supply_power_id",
            "solar_energy": "solar_energy_id",
            "imported_energy": "imported_energy_id",
            "exported_energy": "exported_energy_id",
        },
        session_storage,
        device_type_registry,
    )
    device = HomeassistantDevice(
        {
            "id": "1a8ac2d6-5695-427a-a3c5-ef567b34e5ec",
            "name": "Device 1",
            "power": "power_id",
            "energy": "energy_id",
        },
        session_storage,
        device_type_registry,
    )
    home.add_device(device)
    db = Database()
    home_measurement = await HomeMeasurement.read_last(session, True)
    assert home_measurement is not None
    await db.restore_home_state(home, session)
    assert home.consumed_energy == 125
    assert home.consumed_solar_energy == 121
    assert home.devices[0].consumed_solar_energy == 2
    assert home.devices[0].consumed_energy == 4

    assert True



"""Storage of the Home state including the connected devices."""

from datetime import date, datetime, timezone
import logging
import uuid

from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker

from app.db import get_session
from app.devices import PowerModes, Session, SessionStorage
from app.devices.home import Home
from app.devices.registry import DeviceTypeRegistry
from app.devices.utility_meter import UtilityMeter
from app.models.device import (
    Device as DeviceDTO,
    DeviceMeasurement,
    UtilityMeter as UtilityMeterDTO,
)
from app.models.home import HomeMeasurement
from app.models.sessionlog import SessionLogEntry

from .constants import ROOT_LOGGER_NAME

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


async def get_async_session() -> async_sessionmaker:
    """Get a async database session."""
    return await anext(get_session())


class Database:
    """Database for storing home and devie measurement data."""

    def __init__(self) -> None:
        """Create a Database instance."""

    async def restore_home_state(self, home: Home, session: AsyncSession) -> None:
        """Restore the state of the home from the database."""
        if home is not None:
            try:
                home_measurement = await HomeMeasurement.read_last(session, True)
                if home_measurement is not None:
                    home.restore_state(
                        home_measurement.solar_consumed_energy,
                        home_measurement.consumed_energy,
                        home_measurement.solar_produced_energy,
                        home_measurement.grid_imported_energy,
                        home_measurement.grid_exported_energy,
                    )

                    snapshot_measurement = await HomeMeasurement.read_before_date(
                        session, home_measurement.measurement_date, True
                    )
                    if snapshot_measurement is None:
                        snapshot_measurement = home_measurement
                    home.set_snapshot(
                        snapshot_measurement.solar_consumed_energy,
                        snapshot_measurement.consumed_energy,
                        snapshot_measurement.solar_produced_energy,
                        snapshot_measurement.grid_imported_energy,
                        snapshot_measurement.grid_exported_energy,
                    )
                    for device in home.devices:
                        device_measurement = home_measurement.get_device_measurement(device.id)
                        if device_measurement is not None:
                            device.restore_state(
                                device_measurement.solar_consumed_energy,
                                device_measurement.consumed_energy,
                            )
                        device_measurement = snapshot_measurement.get_device_measurement(device.id)
                        if device_measurement is not None:
                            device.set_snapshot(
                                device_measurement.solar_consumed_energy,
                                device_measurement.consumed_energy,
                            )
                await self.restore_utility_meters(session, home)
            except Exception as ex:
                LOGGER.error("Error while restoring state of home", ex)

    async def restore_utility_meters(self, session: AsyncSession, home: Home) -> None:
        """Restore the utility meters."""
        for device in home.devices:
            for utility_meter in device._utility_meters:
                utility_meter_dto = await UtilityMeterDTO.read_by_name(
                    session, utility_meter.name, device.id
                )
                if utility_meter_dto is not None:
                    utility_meter.restore_last_meter_value(utility_meter_dto.last_meter_value)

    async def store_home_state(self, home: Home, session: AsyncSession) -> None:
        """Store the state of the home including all devices."""
        today = date.today()
        try:
            home_measurement = await HomeMeasurement.read_by_date(session, today, True)
            if home_measurement is None:
                home_measurement = await HomeMeasurement.create(
                    session,
                    name=home.name,
                    measurement_date=today,
                    consumed_energy=home.consumed_energy,
                    solar_consumed_energy=home.consumed_solar_energy,
                    solar_produced_energy=home.produced_solar_energy,
                    grid_imported_energy=home.grid_imported_energy,
                    grid_exported_energy=home.grid_exported_energy,
                    device_measurements=[],
                )
                for device in home.devices:
                    device_dto = await DeviceDTO.read_by_id(session, device.id)
                    if device_dto is not None:
                        await DeviceMeasurement.create(
                            session,
                            home_measurement=home_measurement,
                            consumed_energy=device.consumed_energy,
                            solar_consumed_energy=device.consumed_solar_energy,
                            device=device_dto,
                        )

            else:
                await home_measurement.update(
                    session,
                    name=home.name,
                    measurement_date=today,
                    consumed_energy=home.consumed_energy,
                    solar_consumed_energy=home.consumed_solar_energy,
                    solar_produced_energy=home.produced_solar_energy,
                    grid_imported_energy=home.grid_imported_energy,
                    grid_exported_energy=home.grid_exported_energy,
                )
                for device in home.devices:
                    device_measurement = home_measurement.get_device_measurement(device.id)
                    if device_measurement is not None:
                        device_dto = await DeviceDTO.read_by_id(session, device.id)
                        if device_dto is not None:
                            await device_measurement.update(
                                session,
                                home_measurement=home_measurement,
                                consumed_energy=device.consumed_energy,
                                solar_consumed_energy=device.consumed_solar_energy,
                                device=device_dto,
                            )
            for device in home.devices:
                for utility_meter in device._utility_meters:
                    await self.create_or_update_utility_meter(session, device.id, utility_meter)
            await session.flush()
            await session.commit()
        except Exception as ex:
            LOGGER.error("Error while storing state of home", ex)

    async def update_devices(
        self,
        home: Home,
        session: AsyncSession,
        session_storage: SessionStorage,
        device_type_registry: DeviceTypeRegistry,
    ) -> None:
        """Update the devices table in the db based on the configuration."""
        for device in home.devices:
            try:
                device_dto = await DeviceDTO.read_by_id(session, device.id)
                if device_dto is not None:
                    if device_dto.power_mode is not None:
                        device.set_power_mode(PowerModes[device_dto.power_mode.upper()])
                    await device_dto.update(
                        session,
                        device.name,
                        device.icon,
                        device.power_mode,
                        device.type,
                        device.config,
                    )
                else:
                    await DeviceDTO.create(
                        session,
                        device.id,
                        device.name,
                        device.icon,
                        device.power_mode,
                        device.type,
                        device.config,
                    )
                await session.flush()
                await session.commit()

            except Exception as ex:
                LOGGER.error("Error while udpateing the devices of home", ex)
        all_devices = DeviceDTO.read_all(session)
        async for device_dto in all_devices:
            if home.get_device(device_dto.id) is None and device_dto.type is not None:
                home.create_device(
                    device_dto.type, device_dto.get_config(), session_storage, device_type_registry
                )

    async def create_or_update_utility_meter(
        self, session: AsyncSession, device_id: uuid.UUID, utility_meter: UtilityMeter
    ) -> None:
        """Create or update a utility meter for a given device."""
        utility_meter_dto = await UtilityMeterDTO.read_by_name(
            session, utility_meter.name, device_id
        )
        if utility_meter_dto is None:
            await UtilityMeterDTO.create(
                session, device_id, utility_meter.name, utility_meter.last_meter_value
            )
        else:
            await utility_meter_dto.update(
                session, device_id, utility_meter.name, utility_meter.last_meter_value
            )


class DbSessionStorage(SessionStorage):
    """Session storage in the database."""

    async def start_session(
        self,
        device_id: uuid.UUID,
        text: str,
        solar_consumed_energy: float,
        consumed_energy: float,
    ) -> Session:
        """Start a new session."""
        async_session = await get_async_session()
        async with async_session.begin() as session:
            entry = await SessionLogEntry.create(
                session,
                text,
                device_id,
                datetime.now(timezone.utc),
                solar_consumed_energy,
                consumed_energy,
                datetime.now(timezone.utc),
                solar_consumed_energy,
                consumed_energy,
            )
            return Session(entry.id, entry.start, solar_consumed_energy, consumed_energy)

    async def update_session(
        self, id: int, solar_consumed_energy: float, consumed_energy: float
    ) -> None:
        """Update the session with the given id."""
        async_session = await get_async_session()
        async with async_session.begin() as session:
            entry = await SessionLogEntry.read_by_id(session, id)
            if entry is not None:
                await entry.update(
                    session,
                    entry.text,
                    entry.device_id,
                    entry.start,
                    entry.start_solar_consumed_energy,
                    entry.start_consumed_energy,
                    datetime.now(timezone.utc),
                    solar_consumed_energy,
                    consumed_energy,
                )

    async def update_session_energy(
        self, id: int, solar_consumed_energy: float, consumed_energy: float
    ) -> None:
        """Update the session with the given id."""
        async_session = await get_async_session()
        async with async_session.begin() as session:
            entry = await SessionLogEntry.read_by_id(session, id)
            if entry is not None:
                await entry.update(
                    session,
                    entry.text,
                    entry.device_id,
                    entry.start,
                    entry.start_solar_consumed_energy,
                    entry.start_consumed_energy,
                    entry.end,
                    solar_consumed_energy,
                    consumed_energy,
                )


session_storage = DbSessionStorage()



"""Mqtt connection for energy assistant."""

import copy
from datetime import datetime, timezone
import json
import logging
import pathlib
import pickle
from typing import Tuple
import uuid

from emhass import utils  # type: ignore
from emhass.forecast import forecast  # type: ignore
from emhass.machine_learning_forecaster import mlforecaster  # type: ignore
from emhass.optimization import optimization  # type: ignore
from emhass.retrieve_hass import retrieve_hass  # type: ignore
import numpy as np
import pandas as pd
from sklearn.metrics import r2_score  # type: ignore

from app import Optimizer
from app.devices import Location, StateId, StatesRepository
from app.devices.analysis import DataBuffer
from app.devices.home import Home
from app.devices.homeassistant import HOMEASSISTANT_CHANNEL, Homeassistant
from app.models.forecast import ForecastSchema, ForecastSerieSchema

from .constants import ROOT_LOGGER_NAME

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)

SENSOR_POWER_NO_VAR_LOADS = "power_load_no_var_loads"
DEFAULT_HASS_ENTITY_PREFIX = "em"
DEFAULT_COST_FUNC = "profit"
LOAD_FORECAST_MODEL_TYPE = "load_forecast"


class EmhassOptimizer(Optimizer):
    """Optimizer based on Emhass."""

    def __init__(self, data_folder: str, config: dict, hass: Homeassistant) -> None:
        """Create an emhass optimizer instance."""
        self._data_folder: pathlib.Path = pathlib.Path(data_folder)
        self._logger = LOGGER
        self._hass_url: str = hass.url
        if self._hass_url is not None and self._hass_url[-1] != "/":
            self._hass_url = self._hass_url + "/"
        self._hass_token: str = hass.token
        self._location: Location = hass.get_location()

        home_config = config.get("home")
        self._solar_power_id: str | None = None
        if home_config is not None:
            self._solar_power_id = home_config.get("solar_power")
        self._emhass_config: dict | None = config.get("emhass")
        self._power_no_var_loads_id = (
            f"sensor.{DEFAULT_HASS_ENTITY_PREFIX}_{SENSOR_POWER_NO_VAR_LOADS}"
        )
        if self._emhass_config is not None:
            self._cost_fun = self._emhass_config.get("costfun", DEFAULT_COST_FUNC)
            self._hass_entity_prefix = self._emhass_config.get(
                "hass_entity_prefix", DEFAULT_HASS_ENTITY_PREFIX
            )
            self._power_no_var_loads_id = (
                f"sensor.{self._hass_entity_prefix}_{SENSOR_POWER_NO_VAR_LOADS}"
            )
            params = json.dumps(self._emhass_config)
            retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(
                pathlib.Path(), False, params=params
            )
            # Patch variables with Energy Assistant Config
            retrieve_hass_conf["hass_url"] = self._hass_url
            retrieve_hass_conf["long_lived_token"] = self._hass_token
            retrieve_hass_conf["var_PV"] = self._solar_power_id
            if "var_load" not in retrieve_hass_conf:
                retrieve_hass_conf["var_load"] = self._power_no_var_loads_id
            else:
                self._power_no_var_loads_id = retrieve_hass_conf["var_load"]
            retrieve_hass_conf["var_replace_zero"] = [self._solar_power_id]
            retrieve_hass_conf["var_interp"] = [
                self._solar_power_id,
                self._power_no_var_loads_id,
            ]

            retrieve_hass_conf["time_zone"] = self._location.get_time_zone()
            retrieve_hass_conf["lat"] = self._location.latitude
            retrieve_hass_conf["lon"] = self._location.longitude
            retrieve_hass_conf["alt"] = self._location.elevation

            optim_conf["num_def_loads"] = 0

            self._retrieve_hass_conf = retrieve_hass_conf
            self._optim_conf = optim_conf
            self._plant_conf = plant_conf

            self._method_ts_round = retrieve_hass_conf.get("method_ts_round")

            # Define main objects
            self._retrieve_hass = retrieve_hass(
                self._hass_url,
                self._hass_token,
                retrieve_hass_conf["freq"],
                self._location.get_time_zone(),
                params,
                self._data_folder,
                self._logger,
                get_data_from_file=False,
            )  # type: ignore

        if self._cost_fun is None:
            self._cost_fun = "profit"
        if self._method_ts_round is None:
            self._method_ts_round = "nearest"

        self._day_ahead_forecast: pd.DataFrame | None = None
        self._optimzed_devices: list = []
        self._pv: DataBuffer = DataBuffer()
        self._no_var_loads: DataBuffer = DataBuffer()

    def update_repository_states(self, home: Home, state_repository: StatesRepository) -> None:
        """Calculate the power of the non varibale/non controllable loads."""
        power = home.home_consumption_power
        for device in home.devices:
            if device.power_controllable:
                power = power - device.power
        if power < 0:
            power = 0.0
        self._no_var_loads.add_data_point(power)
        attributes = {
            "unit_of_measurement": "W",
            "state_class": "measurement",
            "device_class": "power",
        }
        state_repository.set_state(
            StateId(id=self._power_no_var_loads_id, channel=HOMEASSISTANT_CHANNEL),
            str(power),
            attributes,
        )
        state_repository.set_state(
            StateId(id=f"sensor.{self._hass_entity_prefix}_p_pv", channel=HOMEASSISTANT_CHANNEL),
            str(self._get_forecast_value("P_PV")),
            attributes,
        )
        state_repository.set_state(
            StateId(
                id=f"sensor.{self._hass_entity_prefix}_p_consumption", channel=HOMEASSISTANT_CHANNEL
            ),
            str(self._get_forecast_value("P_Load")),
            attributes,
        )

    def perfect_forecast_optim(
        self, save_data_to_file: bool = True, debug: bool = False
    ) -> pd.DataFrame:
        """Perform a call to the perfect forecast optimization routine.

        :param input_data_dict:  A dictionnary with multiple data used by the action functions
        :type input_data_dict: dict
        :param logger: The passed logger object
        :type logger: logging object
        :param save_data_to_file: Save optimization results to CSV file
        :type save_data_to_file: bool, optional
        :param debug: A debug option useful for unittests
        :type debug: bool, optional
        :return: The output data of the optimization
        :rtype: pd.DataFrame

        """
        self._logger.info("Setting up needed data")

        # Treat runtimeparams
        params: str = ""
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            None,
            json.dumps(self._emhass_config),
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            "perfect-optim",
            self._logger,
        )  # type: ignore
        fcst = forecast(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            params,
            str(self._data_folder),
            self._logger,
            get_data_from_file=False,
        )
        opt = optimization(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            fcst.var_load_cost,
            fcst.var_prod_price,
            self._cost_fun,
            str(self._data_folder),
            self._logger,
        )

        days_list = utils.get_days_list(self._retrieve_hass_conf["days_to_retrieve"])
        var_list = [self._solar_power_id, self._power_no_var_loads_id]
        self._retrieve_hass.get_data(
            days_list, var_list, minimal_response=False, significant_changes_only=False
        )
        self._retrieve_hass.prepare_data(
            self._retrieve_hass_conf["var_load"],
            load_negative=self._retrieve_hass_conf["load_negative"],
            set_zero_min=self._retrieve_hass_conf["set_zero_min"],
            var_replace_zero=self._retrieve_hass_conf["var_replace_zero"],
            var_interp=self._retrieve_hass_conf["var_interp"],
        )
        df_input_data = self._retrieve_hass.df_final.copy()

        self._logger.info("Performing perfect forecast optimization")
        # Load cost and prod price forecast
        df_input_data = fcst.get_load_cost_forecast(
            df_input_data, method=fcst.optim_conf["load_cost_forecast_method"]
        )
        df_input_data = fcst.get_prod_price_forecast(
            df_input_data, method=fcst.optim_conf["prod_price_forecast_method"]
        )
        opt_res = opt.perform_perfect_forecast_optim(df_input_data, days_list)
        # Save CSV file for analysis
        if save_data_to_file:
            filename = f"opt_res_perfect_optim_{self._cost_fun}.csv"
        else:  # Just save the latest optimization results
            filename = "opt_res_perfect_optim_latest.csv"
        if not debug:
            opt_res.to_csv(self._data_folder / filename, index_label="timestamp")
        return opt_res

    def get_ml_runtime_params(self) -> dict:
        """Get the emhass runtime params for the machine learning load prediction."""
        freq = self._retrieve_hass_conf["freq"].total_seconds() / 3600

        runtimeparams: dict = {
            "num_def_loads": len(self._optimzed_devices),
            "P_deferrable_nom": [device.nominal_power for device in self._optimzed_devices],
            "def_total_hours": [device.deferrable_hours for device in self._optimzed_devices],
            "treat_def_as_semi_cont": [
                not device.is_continous for device in self._optimzed_devices
            ],
            "set_def_constant": [device.is_constant for device in self._optimzed_devices],
            "days_to_retrieve": self._retrieve_hass_conf.get("days_to_retrieve", 10),
            "model_type": LOAD_FORECAST_MODEL_TYPE,
            "var_model": self._power_no_var_loads_id,
            "sklearn_model": "KNeighborsRegressor",
            "num_lags": int(24 / freq),  # should be one day * 30 min
            "split_date_delta": "48h",
            "perform_backtest": True,
        }
        return runtimeparams

    def dayahead_forecast_optim(self, save_data_to_file: bool = False, debug: bool = False) -> None:
        """Perform a call to the day-ahead optimization routine.

        :param save_data_to_file: Save optimization results to CSV file
        :type save_data_to_file: bool, optional
        :param debug: A debug option useful for unittests
        :type debug: bool, optional

        """
        self._logger.info("Setting up needed data for a day ahead forecast")

        # Fall back to the naive method in case the model for the mlforecaster does not exist
        filename_path = self._data_folder / "load_forecast_mlf.pkl"
        if not filename_path.exists():
            self._optim_conf["load_forecast_method"] = "naive"
            self._logger.warning("Falling back to the naive load forecaster.")

        # Treat runtimeparams
        params: str = ""
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            json.dumps(self.get_ml_runtime_params()),
            json.dumps(self._emhass_config),
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            "dayahead-optim",
            self._logger,
        )  # type: ignore
        fcst = forecast(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            params,
            str(self._data_folder),
            self._logger,
            get_data_from_file=False,
        )
        opt = optimization(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            fcst.var_load_cost,
            fcst.var_prod_price,
            self._cost_fun,
            str(self._data_folder),
            self._logger,
        )

        df_weather = fcst.get_weather_forecast(method=self._optim_conf["weather_forecast_method"])
        P_PV_forecast = fcst.get_power_from_weather(df_weather)
        try:
            P_load_forecast = fcst.get_load_forecast(
                method=self._optim_conf["load_forecast_method"]
            )
            P_load_forecast_values = np.array(P_load_forecast.values)
        except Exception:
            self._logger.warning(
                "Forcasting the load failed, probably due to missing history data in Home Assistant."
            )
            avg_non_var_power = self._no_var_loads.average()
            P_load_forecast = pd.Series(
                [avg_non_var_power for x in P_PV_forecast.values],
                index=P_PV_forecast.index,
            )
            P_load_forecast_values = np.array([avg_non_var_power for x in P_PV_forecast.values])

        df_input_data_dayahead = pd.DataFrame(
            np.transpose(np.vstack([np.array(P_PV_forecast.values), P_load_forecast_values])),
            index=P_PV_forecast.index,
            columns=["P_PV_forecast", "P_load_forecast"],
        )
        df_input_data_dayahead = utils.set_df_index_freq(df_input_data_dayahead)

        # params_dayahead: dict = json.loads(params)
        # if 'prediction_horizon' in params_dayahead['passed_data'] and params_dayahead['passed_data']['prediction_horizon'] is not None:
        #    prediction_horizon = params_dayahead['passed_data']['prediction_horizon']
        #    df_input_data_dayahead = copy.deepcopy(df_input_data_dayahead)[df_input_data_dayahead.index[0]:df_input_data_dayahead.index[prediction_horizon-1]]

        self._logger.info("Performing day-ahead forecast optimization")
        # Load cost and prod price forecast
        df_input_data_dayahead = fcst.get_load_cost_forecast(
            df_input_data_dayahead, method=fcst.optim_conf["load_cost_forecast_method"]
        )
        df_input_data_dayahead = fcst.get_prod_price_forecast(
            df_input_data_dayahead, method=fcst.optim_conf["prod_price_forecast_method"]
        )
        self._day_ahead_forecast = opt.perform_dayahead_forecast_optim(
            df_input_data_dayahead, P_PV_forecast, P_load_forecast
        )
        # Save CSV file for publish_data
        if save_data_to_file:
            today = datetime.now(timezone.utc).replace(hour=0, minute=0, second=0, microsecond=0)
            filename = "opt_res_dayahead_" + today.strftime("%Y_%m_%d") + ".csv"
        else:  # Just save the latest optimization results
            filename = "opt_res_latest.csv"
        if not debug and self._day_ahead_forecast is not None:
            self._day_ahead_forecast.to_csv(self._data_folder / filename, index_label="timestamp")

    def naive_mpc_optim(self, save_data_to_file: bool = False, debug: bool = False) -> pd.DataFrame:
        """Perform a call to the naive Model Predictive Controller optimization routine.

        :param input_data_dict:  A dictionnary with multiple data used by the action functions
        :type input_data_dict: dict
        :param logger: The passed logger object
        :type logger: logging object
        :param save_data_to_file: Save optimization results to CSV file
        :type save_data_to_file: bool, optional
        :param debug: A debug option useful for unittests
        :type debug: bool, optional
        :return: The output data of the optimization
        :rtype: pd.DataFrame

        """

        self._logger.info("Setting up needed data")

        runtimeparams: dict = {
            "prediction_horizon": 48  # How many 30 min slots do we predict -> 24h
        }

        # Treat runtimeparams
        params: str = ""
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            json.dumps(runtimeparams),
            json.dumps(self._emhass_config),
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            "naive-mpc-optim",
            self._logger,
        )  # type: ignore
        fcst = forecast(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            params,
            str(self._data_folder),
            self._logger,
            get_data_from_file=False,
        )
        opt = optimization(
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            fcst.var_load_cost,
            fcst.var_prod_price,
            self._cost_fun,
            str(self._data_folder),
            self._logger,
        )

        # Retrieve data from hass
        days_list = utils.get_days_list(1)
        var_list = [self._solar_power_id, self._power_no_var_loads_id]
        self._retrieve_hass.get_data(
            days_list, var_list, minimal_response=False, significant_changes_only=False
        )
        self._retrieve_hass.prepare_data(
            self._retrieve_hass_conf["var_load"],
            load_negative=self._retrieve_hass_conf["load_negative"],
            set_zero_min=self._retrieve_hass_conf["set_zero_min"],
            var_replace_zero=self._retrieve_hass_conf["var_replace_zero"],
            var_interp=self._retrieve_hass_conf["var_interp"],
        )
        df_input_data = self._retrieve_hass.df_final.copy()

        # Get PV and load forecasts
        df_weather = fcst.get_weather_forecast(method=self._optim_conf["weather_forecast_method"])
        P_PV_forecast = fcst.get_power_from_weather(
            df_weather, set_mix_forecast=True, df_now=df_input_data
        )
        P_load_forecast = fcst.get_load_forecast(
            method=self._optim_conf["load_forecast_method"],
            set_mix_forecast=True,
            df_now=df_input_data,
        )
        df_input_data_dayahead = pd.concat(
            [
                pd.Series(P_PV_forecast, name="P_PV_forecast"),
                pd.Series(P_load_forecast, name="P_load_forecast"),
            ],
            axis=1,
        )
        df_input_data_dayahead = utils.set_df_index_freq(df_input_data_dayahead)

        # params_naive_mpc_optim: dict = json.loads(params)
        # if 'prediction_horizon' in params_naive_mpc_optim['passed_data'] and params_naive_mpc_optim['passed_data']['prediction_horizon'] is not None:
        #    prediction_horizon = params_naive_mpc_optim['passed_data']['prediction_horizon']
        #    df_input_data_dayahead = copy.deepcopy(df_input_data_dayahead)[df_input_data_dayahead.index[0]:df_input_data_dayahead.index[prediction_horizon-1]]

        self._logger.info("Performing naive MPC optimization")
        # Load cost and prod price forecast
        df_input_data_dayahead = fcst.get_load_cost_forecast(
            df_input_data_dayahead, method=fcst.optim_conf["load_cost_forecast_method"]
        )
        df_input_data_dayahead = fcst.get_prod_price_forecast(
            df_input_data_dayahead, method=fcst.optim_conf["prod_price_forecast_method"]
        )

        # The specifics params for the MPC at runtime
        # TODO: Make this real parameters
        params_dict = json.loads(params)
        prediction_horizon = params_dict["passed_data"]["prediction_horizon"]
        soc_init = params_dict["passed_data"]["soc_init"]
        soc_final = params_dict["passed_data"]["soc_final"]
        def_total_hours = [
            1,
            1,
        ]  #  input_data_dict['params']['passed_data']['def_total_hours']

        opt_res_naive_mpc = opt.perform_naive_mpc_optim(
            df_input_data_dayahead,
            P_PV_forecast,
            P_load_forecast,
            prediction_horizon,
            soc_init,
            soc_final,
            def_total_hours,
        )
        # Save CSV file for publish_data
        if save_data_to_file:
            today = datetime.now(timezone.utc).replace(hour=0, minute=0, second=0, microsecond=0)
            filename = "opt_res_naive_mpc_" + today.strftime("%Y_%m_%d") + ".csv"
        else:  # Just save the latest optimization results
            filename = "opt_res_naive_mpc_latest.csv"
        if not debug:
            opt_res_naive_mpc.to_csv(self._data_folder / filename, index_label="timestamp")
        return opt_res_naive_mpc

    def forecast_model_fit(
        self, only_if_file_does_not_exist: bool = False, debug: bool = False
    ) -> float:
        """Perform a forecast model fit from training data retrieved from Home Assistant.

        :param debug: True to debug, useful for unit testing, defaults to False
        :type debug: Optional[bool], optional
        :return: The DataFrame containing the forecast data results without and with backtest and the `mlforecaster` object
        :rtype: Tuple[pd.DataFrame, pd.DataFrame, mlforecaster]
        """

        filename = LOAD_FORECAST_MODEL_TYPE + "_mlf.pkl"
        filename_path = self._data_folder / filename

        if only_if_file_does_not_exist and filename_path.is_file():
            self._logger.info("Skip model creation")
            return 0

        self._logger.info("Setting up needed data")

        # Treat runtimeparams
        params: str = ""
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            json.dumps(self.get_ml_runtime_params()),
            json.dumps(self._emhass_config),
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            "forecast-model-fit",
            self._logger,
        )  # type: ignore

        params_dict: dict = json.loads(params)
        # Retrieve data from hass
        days_to_retrieve = self._retrieve_hass_conf.get("days_to_retrieve", 10)

        days_list = utils.get_days_list(days_to_retrieve)
        var_list = [self._power_no_var_loads_id]
        self._retrieve_hass.get_data(days_list, var_list)
        df_input_data = self._retrieve_hass.df_final.copy()

        data = copy.deepcopy(df_input_data)
        model_type = params_dict["passed_data"]["model_type"]
        sklearn_model = params_dict["passed_data"]["sklearn_model"]
        num_lags = params_dict["passed_data"]["num_lags"]
        split_date_delta = params_dict["passed_data"]["split_date_delta"]
        perform_backtest = params_dict["passed_data"]["perform_backtest"]
        # The ML forecaster object
        mlf = mlforecaster(
            data,
            model_type,
            self._power_no_var_loads_id,
            sklearn_model,
            num_lags,
            str(self._data_folder),
            self._logger,
        )
        # Fit the ML model
        df_pred, df_pred_backtest = mlf.fit(
            split_date_delta=split_date_delta, perform_backtest=perform_backtest
        )
        predictions = df_pred["pred"].dropna()
        test_data = df_pred["test"].dropna()
        r2 = r2_score(test_data, predictions)
        self._logger.info(f"R2 score = {r2}")
        # Save model
        if not debug:
            with open(self._data_folder / filename, "wb") as outp:
                pickle.dump(mlf, outp, pickle.HIGHEST_PROTOCOL)
        return r2

    def forecast_model_tune(self) -> Tuple[pd.DataFrame, mlforecaster]:
        """Tune a forecast model hyperparameters using bayesian optimization.

        :param debug: True to debug, useful for unit testing, defaults to False
        :type debug: bool, optional
        :return: The DataFrame containing the forecast data results using the optimized model
        :rtype: pd.DataFrame
        """
        # Load model
        self._logger.info("Tune the forecast model")

        filename = LOAD_FORECAST_MODEL_TYPE + "_mlf.pkl"
        filename_path = self._data_folder / filename
        if filename_path.is_file():
            with open(filename_path, "rb") as inp:
                mlf = pickle.load(inp)
            # Tune the model
            df_pred_optim = mlf.tune(debug=False)

            # Save model
            with open(filename_path, "wb") as outp:
                pickle.dump(mlf, outp, pickle.HIGHEST_PROTOCOL)
            return df_pred_optim, mlf

        else:
            self._logger.error(
                "The ML forecaster file was not found, please run a model fit method before this tune method"
            )
            raise Exception(
                "The ML forecaster file was not found, please run a model fit method before this tune method"
            )

    def forecast_model_predict(
        self,
        use_last_window: bool = True,
        debug: bool = False,
        mlf: mlforecaster | None = None,
    ) -> pd.Series | None:
        """Perform a forecast model predict using a previously trained skforecast model.

        :param input_data_dict: A dictionnary with multiple data used by the action functions
        :type input_data_dict: dict
        :param logger: The passed logger object
        :type logger: logging.Logger
        :param use_last_window: True if the 'last_window' option should be used for the \
            custom machine learning forecast model. The 'last_window=True' means that the data \
            that will be used to generate the new forecast will be freshly retrieved from \
            Home Assistant. This data is needed because the forecast model is an auto-regressive \
            model with lags. If 'False' then the data using during the model train is used. Defaults to True
        :type use_last_window: Optional[bool], optional
        :param debug: True to debug, useful for unit testing, defaults to False
        :type debug: Optional[bool], optional
        :param mlf: The 'mlforecaster' object previously trained. This is mainly used for debug \
            and unit testing. In production the actual model will be read from a saved pickle file. Defaults to None
        :type mlf: Optional[mlforecaster], optional
        :return: The DataFrame containing the forecast prediction data
        :rtype: pd.DataFrame
        """
        # Treat runtimeparams
        params: str = ""
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            json.dumps(self.get_ml_runtime_params()),
            json.dumps(self._emhass_config),
            self._retrieve_hass_conf,
            self._optim_conf,
            self._plant_conf,
            "forecast-model-fit",
            self._logger,
        )  # type: ignore

        params_dict: dict = json.loads(params)
        # Retrieve data from hass
        days_to_retrieve = params_dict["passed_data"]["days_to_retrieve"]

        days_list = utils.get_days_list(days_to_retrieve)
        var_list = [self._power_no_var_loads_id]
        self._retrieve_hass.get_data(days_list, var_list)
        df_input_data = self._retrieve_hass.df_final.copy()

        # Load model
        model_type = "load_forecast"
        filename = model_type + "_mlf.pkl"
        filename_path = self._data_folder / filename
        if not debug:
            if filename_path.is_file():
                with open(filename_path, "rb") as inp:
                    mlf = pickle.load(inp)
            else:
                self._logger.error(
                    "The ML forecaster file was not found, please run a model fit method before this predict method"
                )
                return None
        # Make predictions
        if use_last_window:
            data_last_window = copy.deepcopy(df_input_data)
        else:
            data_last_window = None
        if mlf is not None:
            return mlf.predict(data_last_window)
        return None

    def get_forecast(self) -> ForecastSchema:
        """Get the previously calculated forecast."""
        if self._day_ahead_forecast is not None:
            freq = self._retrieve_hass_conf["freq"]
            temp_folder = self._data_folder / "temp"
            temp_folder.mkdir(parents=True, exist_ok=True)
            pv_df = self._pv.get_data_frame(freq, self._location.get_time_zone(), "pv", temp_folder)
            pv_df.to_csv(temp_folder / "pv_df.csv")
            no_var_load_df = self._no_var_loads.get_data_frame(
                freq, self._location.get_time_zone(), "non_var_loads", temp_folder
            )
            df = self._day_ahead_forecast.merge(
                pv_df, how="left", left_index=True, right_index=True
            )
            df = df.merge(no_var_load_df, how="left", left_index=True, right_index=True)

            df.rename(columns={"P_PV": "pv_forecast"}, inplace=True)
            df.to_csv(temp_folder / "forecast.csv", index_label="time_stamp")

            while not pd.notnull(df["pv_forecast"][0]) and len(df.index) > 0:
                df.drop(df.index[0], inplace=True)

            pv_series = [x for x in df["pv"].to_list() if pd.notnull(x)]
            no_var_load_series = [x for x in df["non_var_loads"].to_list() if pd.notnull(x)]

            # TODO: Should be removed
            df.fillna(-10000, inplace=True)
            time_series = df.index.to_series()
            time: list[datetime] = time_series.tolist()
            pv_forecast = df["pv_forecast"].to_list()
            load = df["P_Load"].to_list()

            series = [
                ForecastSerieSchema(name="pv_forecast", data=pv_forecast),
                ForecastSerieSchema(name="pv", data=pv_series),
                ForecastSerieSchema(name="consumption", data=load),
                ForecastSerieSchema(name="no_var_loads", data=no_var_load_series),
            ]
            for i, d in enumerate(self._optimzed_devices):
                device = self._day_ahead_forecast[f"P_deferrable{i}"].to_list()
                series.append(ForecastSerieSchema(name=str(d.device_id), data=device))
            return ForecastSchema(time=time, series=series)
        else:
            raise Exception("Optimizer forecast is not initialized.")

    def get_optimized_power(self, device_id: uuid.UUID) -> float:
        """Get the optimized power budget for a give device."""
        if self._day_ahead_forecast is not None:
            for i, deferrable_load_info in enumerate(self._optimzed_devices):
                if deferrable_load_info.device_id == device_id:
                    columnName = f"P_deferrable{i}"
                    return self._get_forecast_value(columnName)
        return -1

    def _get_forecast_value(self, columnName: str) -> float:
        """Get a forcasted value."""
        if self._day_ahead_forecast is not None:
            now_precise = datetime.now(self._location.get_time_zone()).replace(
                second=0, microsecond=0
            )
            if self._method_ts_round == "nearest":
                method = "nearest"
            elif self._method_ts_round == "first":
                method = "ffill"
            elif self._method_ts_round == "last":
                method = "bfill"
            else:
                method = "nearest"

            idx_closest = self._day_ahead_forecast.index.get_indexer([now_precise], method=method)[  # type: ignore
                0
            ]
            if idx_closest == -1:
                idx_closest = self._day_ahead_forecast.index.get_indexer(  # type: ignore
                    [now_precise], method="nearest"
                )[0]

            value = self._day_ahead_forecast.iloc[idx_closest][columnName]
            return float(value)
        return -1

    def _has_deferrable_load(self, device_id: uuid.UUID) -> bool:
        for deferrable_load_info in self._optimzed_devices:
            if deferrable_load_info.device_id == device_id:
                return True
        return False

    def update_devices(self, home: Home) -> None:
        """Update the selected devices from the list of devices."""
        new_optimizhed_devices = []
        needs_update = False
        self._pv.add_data_point(home.solar_production_power)
        for device in home.devices:
            deferrable_load_info = device.get_deferrable_load_info()
            if deferrable_load_info is not None:
                if not self._has_deferrable_load(device.id):
                    needs_update = True
                new_optimizhed_devices.append(deferrable_load_info)
            else:
                if self._has_deferrable_load(device.id):
                    needs_update = True
        self._optimzed_devices = new_optimizhed_devices
        if needs_update:
            self.dayahead_forecast_optim()
